/* local.h: Research Triangle Institute 4.2bsd ("rti") version */
/*
 * Some systems require ctype.h to get "isdigit(c)", etc., some don't.
 */
#ifndef	isdigit
#include <ctype.h>
#endif

/*
 * The "sitename()" routine in local.c returns the local system name.
 * If you want a special #define to control what code is generated
 * for that routine, put it here.
 * The following are standard(!):
 * UNAME     -- A possibly standard version of uname()
 * ETCWHOAMI -- Read system name from /etc/whoami
 * GETWHOAMI -- Read system name as #defined in the file <whoami.h>
 * CCWHOAMI  -- Compile the #define in <whoami.h> into news. Fast,
 *			but the binary cannot be used across machines.
 */
/*#define		UNAME*/
/*#define	ETCWHOAMI*/
/*#define	GETWHOAMI*/
#define	CCWHOAMI

/*
 * If your system has the "getdate" routine,
 * #define GETDATE below.  Otherwise comment it out.
 * Some things do not work correctly without GETDATE.
 */
#define	GETDATE

/*
 * If your system is to store articles in subdirectories,
 * #define SUBDIR, below.  Otherwise comment it out.
 * To convert a pre-subdir news, run "dirconvert"
 * before this version is installed.
 */
#define	SUBDIR	100

/*
 * CNID is a comma separated *and* terminated list of uids
 * who are permitted to cancel any articles.
 * An article may be cancelled by the submittor
 * or by one of the people mentioned in CNID.
 * CNID entries may be numberic uids or login names.
 */
#define	CNID	"rbw,trt,dlf,lrl,swd,jte,crt,dbl,news,daemon,0,"

/*
 * If _IORW is defined, then "r+" fopen() mode is assumed to work.
 * This permits read/write files in stdio, and improves efficiency.
 * For portability, however, the #define RDWR should be commented out.
 * NOTE 4.1bsd read/write code has bugs.  DO NOT USE.
 */
#define	_IORWBUG

#ifndef	_IORWBUG
#ifdef	_IORW
#define	USE_RDWR
#endif
#endif

/*
 * ARTSIZ is the maximum number of articles
 * that can be processed at one time by news.
 * The storage requirement is ARTSIZ*20 bytes in this version.
 * (This storage is also used by makebits().)
 * Approximate maxima: 300 for non i+d PDP11, 1500 for i+d PDP11,
 * and as big as you want (say 5000) for a VAX.
 */
#define	ARTSIZ	5000

/*
 * SORTNINDEX is a command used to resort the nindex file
 * to ensure proper ordering by time.
 * The command is run with news' powers in the NEWSDIR directory.
 */
#define	SORTNINDEX	"sort -t: -n +1 nindex -o nindex"

/*
 * If LOG is defined, then news will maintain a log file
 * of interesting happenings.
 */
#define	LOG

/*
 * Long articles can be piped to a program such as Berkeley's "more".
 * If you want a default PAGER, #define PAGER below.
 * The 'PAGER' environment variable overrides the default.
 */
#define	PAGER	"/usr/ucb/more"

/*
 * News occasionally needs to mail messages to people.
 * It understands two types of mailers; those which
 * can take a '-s' argument to specify a subject,
 * and those which cannot.  If you have the first
 * type, then SMAILER should be set to its name.
 * Otherwise define NMAILER.  Exactly one of SMAILER
 * and NMAILER must be defined.
 */
#define	SMAILER	"Mail"
/*#define	NMAILER	"mail"*/

/*
 * These #defines control the conditional compilation
 * of the code which handles network control messages.
 * If you are not part of USENET, you should not define
 * CTL_MSG, removing *all* code to handle control messages.
 * This makes news smaller.
 */
#define	CTL_MSG

/*
 * CTL_BUGOFF: the senduuname request is supposed to mail the
 * output of uuname to the requestor.  If CTL_BUGOFF
 * is defined, it just sends a short note denying
 * the requestor such information.  It is USENET
 * policy that some reponse be generated to this
 * request.  The minimal acceptable response is that
 * generated by the CTL_BUGOFF code.
 */
#ifdef CTL_MSG
/*#define	CTL_BUGOFF*/
#endif

/*
 * Most UNIX systems do not honor the setuid bit
 * when root invokes a setuid program.
 * If your system has this bug, SUIDBUG should be defined.
 * If your system *always* honors the setuid,
 * you can comment this out and make news smaller.
 */
/*#define	SUIDBUG*/

/*
 * News runs other programs (e.g. uux) on its behalf
 * restricting the PATH searched for security reasons.
 * The following defines the PATH to be used.
 */
#define	SYSPATH	"PATH=/bin:/usr/bin"

/*
 * The number of uids supported by news is limited by USIZE.
 * News can handle uids from 0 up to USIZE-1.
 * USIZE can be made larger (as large as 32000, if need be),
 * but with a loss of efficiency.  Every 4096 uids take a disk block.
 */
#define	USIZE	4000

/*
 * If PW_ADD is defined, then uids in /etc/passwd but not in uindex
 * will be added to uindex.
 * This means everyone automatically receives ADMSUB news.
 * If you want receiving news to be a purely personal decision,
 * comment out the #define PW_ADD.
 */
#define	PW_ADD

/*
 * If PW_DEL is defined, then uids in uindex but not in /etc/passwd
 * will be dropped from uindex.
 * This provides an automatic cleanup of deleted users.
 * If you do not want this, comment out the #define PW_DEL.
 */
#define	PW_DEL
